import re

from vaurien.protocols.base import BaseProtocol
from vaurien.util import chunked


RE_LEN = re.compile('Content-Length: (\d+)', re.M | re.I)
RE_KEEPALIVE = re.compile('Connection: Keep-Alive')
RE_MEMCACHE_COMMAND = re.compile('(.*)\r\n')

EOH = '\r\n\r\n'
CRLF = '\r\n'


class Redis(BaseProtocol):
    """Redis protocol.
    """
    name = 'redis'

    def _handle(self, source, dest, to_backend):
        """ see http://redis.io/topics/protocol
        """
        # Getting the request.
        buffer = self._get_data(source)
        if not buffer:
            self._abort_handling(to_backend, dest)
            return False

        # Sending the request to the backend.
        dest.sendall(buffer)

        # Getting the answer back and sending it over.
        buffer_size = self.option('buffer')
        buffer = dest.recv(buffer_size)
        source.sendall(buffer)
        already_read = len(buffer)

        if buffer[0] in ('+', '-', ':'):
            # simple reply, we're good
            return False    # disconnect mode ?

        if buffer[0] == '$':
            # bulk reply
            size = int(buffer[1:buffer.find(CRLF)])
            left_to_read = size - already_read

            if left_to_read > 0:
                for chunk in chunked(left_to_read, buffer_size):
                    data = source.recv(chunk)
                    buffer += data
                    dest.sendall(data)

            return False  # disconnect mode ?

        if buffer[0] == '*':
            # multi-bulk reply
            raise NotImplementedError()

        raise NotImplementedError()



from unittest import TestCase
import json

from vaurien.webserver import create_app
from vaurien.tests.support import FakeProxy


class TestHTTPServer(TestCase):
    def setUp(self):
        self.app = create_app()
        self.app.config['TESTING'] = True
        setattr(self.app, 'proxy', FakeProxy())
        self.client = self.app.test_client()

    def tearDown(self):
        self.app = None

    def assertJSONEquals(self, res, value, status=200):
        self.assertEquals(res.headers['Content-Type'], 'application/json')
        self.assertEquals(res.status_code, status)
        self.assertEquals(json.loads(res.data), value)

    def test_behaviors_are_returned(self):
        # the behaviors should be returned as a list of behaviors, in json.
        res = self.client.get('/behaviors')
        self.assertJSONEquals(res, {'behaviors': ['default', 'blackout']})

    def test_get_behavior(self):
        # test that we can know what is the behavior currently used
        res = self.client.get('/behavior')
        self.assertJSONEquals(res, {'behavior': self.app.proxy.behavior})

    def test_set_unknown_behavior(self):
        # we shuld get errors back from the server if we try to set a behavior
        # which doesn't exist
        behavior = {'name': 'doesnt_exist'}
        res = self.client.post('/behavior', data=json.dumps(behavior))
        self.assertEquals(res.headers['Content-Type'], 'application/json')
        self.assertEquals(res.status_code, 400)
        errors = json.loads(res.data)
        self.assertEquals(errors['status'], 'error')
        self.assertEquals(len(errors['errors']), 1)
        self.assertEquals(errors['errors'][0]['location'], 'body')
        self.assertEquals(errors['errors'][0]['name'], 'name')

    def test_set_valid_behavior_name(self):
        # it's possible to only set a behavior with its name
        behavior_name = self.app.proxy.behaviors[-1]
        behavior = {'name': behavior_name}
        res = self.client.post('/behavior', data=json.dumps(behavior))
        self.assertJSONEquals(res, {'status': 'ok'})
        self.assertEquals(self.app.proxy.behavior, behavior_name)

    def test_set_valid_behavior_with_options(self):
        # it's possible to pass options with the behavior
        behavior_name = self.app.proxy.behaviors[-1]
        behavior = {'name': behavior_name}
        behavior_options = {'foo': 'bar', 'baz': 'babar'}
        behavior.update(behavior_options)
        res = self.client.post('/behavior', data=json.dumps(behavior))
        self.assertJSONEquals(res, {'status': 'ok'})
        self.assertEquals(self.app.proxy.behavior, behavior_name)
        self.assertEquals(self.app.proxy.behavior_options, behavior_options)



import json
from contextlib import contextmanager
from urlparse import urlparse

import requests


class Client(object):
    def __init__(self, host='localhost', port=8080, scheme='http'):
        self.host = host
        self.port = port
        self.scheme = scheme
        self.root_url = '%s://%s:%s' % (scheme, host, port)
        self.behavior_url = self.root_url + '/behavior'
        self.list_behaviors_url = self.root_url + '/behaviors'

    def set_behavior(self, behavior, **options):
        options['name'] = behavior
        res = requests.post(self.behavior_url, data=json.dumps(options))
        if res.status_code == 400:
            errors = res.json.get('errors', [])
            status = res.json.get('status', 'error')
            if status == 'error' and errors[0]['name'] == 'name':
                raise ValueError(errors[0]['description'])
        res.raise_for_status()

    def get_behavior(self):
        res = requests.get(self.behavior_url)
        res.raise_for_status()
        return res.json['behavior']

    def list_behaviors(self):
        res = requests.get(self.list_behaviors_url)
        res.raise_for_status()
        return res.json['behaviors']

    @contextmanager
    def with_behavior(self, behavior, **options):
        current = self.get_behavior()
        self.set_behavior(behavior, **options)
        try:
            yield
        finally:
            self.set_behavior(current)


def main():
    """Command-line tool to change the behavior that's being used by vaurien"""
    import argparse
    parser = argparse.ArgumentParser(description='Change the vaurien behavior')
    parser.add_argument('action', help='The action you want to do.',
                        choices=['list-behaviors', 'set-behavior',
                                 'get-behavior'])
    parser.add_argument('behavior', nargs='?',
                        help='The vaurien behavior to set for the next calls')
    parser.add_argument('--host', dest='host', default='http://localhost:8080',
                        help='The host to use. Provide the scheme.')

    args = parser.parse_args()
    parts = urlparse(args.host)
    scheme = parts.scheme
    host, port = parts.netloc.split(':', -1)
    client = Client(host, port, scheme)
    if args.action == 'list-behaviors':
        print ', '.join(client.list_behaviors())
    elif args.action == 'set-behavior':
        try:
            client.set_behavior(args.behavior)
            print 'Behavior changed to "%s"' % args.behavior
        except ValueError:
            print 'The request failed. Please use one of %s' %\
                ', '.join(client.list_behaviors())
    elif args.action == 'get-behavior':
        print client.get_behavior()



from abc import ABCMeta, abstractmethod


class Protocol(object):
    """Registry for protocol.

    A protocol is a class that implements a __call__ method.
    """

    __metaclass__ = ABCMeta
    _cache = {}

    @abstractmethod
    def __call__(self, client_sock, backend_sock, to_backend):
        pass

    @classmethod
    def __subclasshook__(cls, klass):
        if cls is Protocol:
            for method in cls.__abstractmethods__:
                if any(method in base.__dict__ for base in klass.__mro__):
                    continue
                return NotImplemented
            return True
        return NotImplemented

    @classmethod
    def register(cls, subclass):
        ABCMeta.register(cls, subclass)
        if subclass not in cls._abc_registry:
            cls._abc_registry.add(subclass)

    @classmethod
    def _get_instance(cls, klass):
        name = klass.name
        if name not in cls._cache:
            cls._cache[name] = klass()
        return cls._cache[name]

    @classmethod
    def get_protocols(cls):
        return dict([(klass.name, cls._get_instance(klass))
                     for klass in cls._abc_registry])

    @classmethod
    def get_protocol(cls, name):
        for klass in cls._abc_registry:
            if klass.name == name:
                return cls._get_instance(klass)
        raise KeyError(name)


def get_protocols():
    """Return a mapping of all protocols.

    The keys are the protocol names and the values the classes.
    """
    return Protocol.get_protocols()


def get_protocol(name):
    """Returns an instance of the given protocol."""
    return Protocol.get_protocol(name)()


# manually register built-in protocols
from vaurien.protocols.tcp import TCP
Protocol.register(TCP)

from vaurien.protocols.redis import Redis
Protocol.register(Redis)

from vaurien.protocols.memcache import Memcache
Protocol.register(Memcache)

from vaurien.protocols.http import Http
Protocol.register(Http)



"""A simple, flask-based webserver able to control how the proxy behaves"""
import os
import json

try:
    from flask import (Flask, Blueprint, request, request_started,
                       request_finished, jsonify, current_app)
except ImportError as e:
    reqs = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                        'web-requirements.txt')

    raise ImportError('You need some dependencies to run the web interface. '
                      'You can do so by using "pip install -r '
                      '%s"\nInitial error: %s' % (reqs, str(e)))

api = Blueprint('api', __name__)


@api.route('/behaviors', methods=['GET'])
def get_behaviors():
    return jsonify(behaviors=current_app.proxy.get_behavior_names())


@api.route('/behavior', methods=['POST', 'GET'])
def update_behavior():
    if request.method == 'POST':
        try:
            data = json.loads(request.data)
            name = data['name']
        except ValueError:
            request.errors.add('body', '',
                               'the value is not a valid json object')
        except KeyError:
            request.errors.add('body', '',
                               'the value should contain a "name" key')
        else:
            try:
                current_app.proxy.set_behavior(**data)
            except KeyError:
                request.errors.add('body', 'name',
                                "the '%s' behavior does not exist" % name)

        if len(request.errors) == 0:
            return jsonify(status='ok')
        else:
            return jsonify(status='error')
    else:
        return jsonify(behavior=current_app.proxy.get_behavior()[1])


# utils

class Errors(list):
    """Holds Request errors
    """
    def __init__(self, status=400):
        self.status = status
        super(Errors, self).__init__()

    def add(self, location, name=None, description=None):
        """Registers a new error."""
        self.append(dict(
            location=location,
            name=name,
            description=description))


def add_errors(sender, **extra):
    """Add errors to the request"""
    setattr(request, 'errors', Errors())


def convert_errors(sender, response, **extra):
    """convert errors to json if needed"""
    if len(request.errors) > 0:
        response.data = json.dumps({'status': 'error',
                                    'errors': request.errors})
        response.status_code = request.errors.status
    # as we always answer in json for now...
    response.content_type = 'application/json'


def create_app(debug=False):
    app = Flask(__name__)
    app.register_blueprint(api)

    request_started.connect(add_errors, app)
    request_finished.connect(convert_errors, app)

    if debug:
        app.debug = True
    return app

app = create_app()
